#include<stdio.h>
#include<string>
#include<list>
#include<stack>
#include<string.h>

typedef unsigned char BYTE;


//!  
class XML_core{
	std::string tag,attr,value,body; // 
	int rem_index;	//!<    
	int status;		//!<   
private:
	virtual void _open(const char*);				//!<  -  
	virtual void _text(const char*);				//!<  - 
	virtual void _attr(const char*,const char*);	//!<  - 
	virtual void _close(const char*);				//!<  -  
public:
	XML_core() {ParseStart();}
	void ParseStart();								//!<  
	bool ParseEnd();								//!<  
	void Parse(char sym);							//!<   
	void Parse(char *text);							//!<   
};

//!  
class XML_simple:public XML_core{
	bool isOpen;							//!<      
	std::stack<char*>st;					//!<  
	void _open(const char*);				//!<  -  
	void _text(const char*);				//!<  - 
	void _attr(const char*,const char*);	//!<  - 
	void _close(const char*);				//!<  -  
	void Clear();							//!<  
public:
	XML_simple():isOpen(false){}
	~XML_simple() {Clear();}
	void ParseStart();						//!<  
	bool ParseEnd();						//!<  
};

//!   
class XML_tree:public XML_core{
	bool isOpen;							//!<      
	struct attr_st{
		char *name;		//!<  
		char *value;	//!<  
	};					//!< 
	struct node_st{
		char *name;						//!<  
		node_st *parent;				//!<  
		std::list<attr_st*> attrs;		//!<  
		std::list<node_st*> nodes;		//!<  
	};					//!< 
	node_st *current;					//!<     
	node_st root;						//!< 
	void Clear();						//!<  
	void Clear(node_st *node);			//!<   -  
	void Dump(node_st *node);			//!<   -  
public:
	XML_tree() {
		isOpen=false;
		root.name="";
		current=&root;
	}
	~XML_tree() {
		Clear();
	}
	void ParseStart();						//!<  
	void Dump();							//!<  
	void _open(const char*);				//!<  -  
	void _text(const char*);				//!<  - 
	void _attr(const char*,const char*);	//!<  - 
	void _close(const char*);				//!<  -  
};

//////////////////////////////////////////////////////////////////////////
// Core section

void XML_core::_open(const char* text)
{
	printf("open: %s\n",text);	
}

void XML_core::_text(const char* text)
{
	printf("text: %s\n",text);
}

//! \param name  
//! \param value  
void XML_core::_attr(const char* name,const char* value)
{
	printf("attr: %s=%s\n",name,value);
}

void XML_core::_close(const char* text)
{
	if(!text || !*text) text="*";
	printf("close: %s\n",text);
}

void XML_core::ParseStart()
{
	status=0;
}

//! \result true  XML    
bool XML_core::ParseEnd()
{
	Parse((char)0);
	return !!status;
}


void XML_core::Parse(char *text)
{
	while(*text) Parse(*text++);
}


// 0 - undef
// < -  ,  
// > -   ,   
// = -   ,  '='
// / -       
// v -  '='   
// t -   , 
// ' -    
// ! -   

void XML_core::Parse(char sym)
{
	int nstatus=status;
#if 1	//       
	switch(status) {
	case 0:
		if(sym=='<') nstatus='<'; else
		if((BYTE)sym>' ') nstatus='t';
		break;
	case '<':
		if(sym=='>') nstatus=0; else
		if(sym=='/' && !tag.empty()) nstatus='/'; 
		else if((BYTE)sym<=' ') nstatus='>';
		break;
	case '>':
		if(sym=='>') nstatus=0; else
		if((BYTE)sym>' ') nstatus='=';
		break;
	case 'v':
		if(sym=='\'' || sym=='\"') nstatus=sym; else nstatus='>';
		break;
	case '=':
		if(sym=='>') nstatus=0;
		if(sym=='=') nstatus='v';
		if(sym==' ') nstatus='>';
		break;
	case '\'':
	case '\"':
		if(sym==status) nstatus='>';
		break;
	case 't':
		if(sym=='<') nstatus='<';
		break;
	case '!':
		if(rem_index>=0 && sym!="-->"[rem_index++]) rem_index=-1;
		if(rem_index<0 && sym!='-') rem_index=0; 
		if(rem_index>2) nstatus=0;
		break;
	}
	if(!sym) status=0;
#else	
	if(status=='!') {
		if(rem_index>=0 && sym!="-->"[rem_index++]) rem_index=-1;
		if(rem_index<0 && sym!='-') rem_index=0; 
		if(rem_index>2) nstatus=0;
	} else switch(sym) {
	case 0: 
		nstatus=0;
		break;
	case '/': 
		if(status!='t' && status!='\'' && status!='\"' && !tag.empty()) nstatus='/';
		break;
	case '<': 
		if(!status || status=='t') nstatus='<'; 
		break;
	case '>': 
		if(status!='t' && status!='\'' && status!='\"') nstatus=0; 
		break;
	case '\'':
	case '\"':
		if(status==sym) nstatus='>';
		else if(status=='v') nstatus=sym;
		else if(!status) nstatus='t';
		break;
	case '=': 
		if(status=='=') nstatus='v';
		if(!status) nstatus='t';
		break;
	default:
		if(status=='v') nstatus='>';
		if((BYTE)sym<=' ' && status) nstatus='>';
		if((BYTE)sym>' ' && !status) nstatus='t';
		if((BYTE)sym>' ' && status=='>') nstatus='=';
		break;
	}
#endif
	if(status!=nstatus) {
		switch(status) {
		case '=':
			if(attr=="/") _close(0);
			break;
		case 't': 
			_text(body.c_str());
			break;
		case '\'': 
		case '\"': 
			_attr(attr.c_str(),value.c_str()); break;
		case '<': 
			if(tag=="!--") nstatus='!',rem_index=-1;
			else if(tag.empty()) _close(0);
			else if(tag[0]=='/') _close(tag.c_str()+1);
			else if(tag[0]!='?') _open(tag.c_str());
			break;
		}
		status=nstatus;
		switch(status) {
		case 't': body.erase(); break;
		case '/': status='<';
		case '<': tag.erase(); return;
		case '=': attr.erase(),value.erase(); break;
		case '\'':
		case '\"':
		case 'v': return;
		}
	}
	switch(status) {
	case 't':  body+=sym; break;
	case '<':  tag +=sym; break;
	case '=':  attr+=sym; break;
	case '\'': 
	case '\"': value+=sym; break;
	}
}


//////////////////////////////////////////////////////////////////////////
// Simple section

void XML_simple::_open(const char *name)
{
	isOpen=true;
	st.push(strdup(name));
	printf("open: %s\n",name);
}

void XML_simple::_close(const char *name)
{
	isOpen=false;
	if(st.empty()) {
		printf("empty stack: %s\n",name);
		return;
	}
	if(name && *name) {
		if(strcmp(st.top(),name)) {
			printf("nomath: %s\n",name);
			return;
		}
	}
	printf("close: %s\n",st.top());
	st.pop();
}

void XML_simple::_text(const char* text)
{
	printf("%s\n",text);
}


//! \param name  
//! \param value  
void XML_simple::_attr(const char* name,const char* value)
{
	if(!isOpen) return;
	printf("%s=%s\n",name,value);
}

void XML_simple::Clear()
{
	while(!st.empty()) {
		free(st.top());
		st.pop();
	}
}

void XML_simple::ParseStart()
{
	isOpen=false;
	XML_core::ParseStart();
}

//! \result true  XML    
bool XML_simple::ParseEnd()
{
	return XML_core::ParseEnd()||!st.empty();
}

//////////////////////////////////////////////////////////////////////////
// Tree section


void XML_tree::Dump() 
{
	printf("[dump]\n");
	Dump(&root);
}

void XML_tree::ParseStart()
{
	Clear();
	isOpen=false;
	XML_core::ParseStart();
}

void XML_tree::Clear()
{
	Clear(&root);
}

void XML_tree::Clear(node_st *node)
{
	for(std::list<attr_st*>::iterator it1=node->attrs.begin(); it1!=node->attrs.end(); it1++) {
		free((*it1)->name);
		free((*it1)->value);
		delete *it1;
	}
	for(std::list<node_st*>::iterator it2=node->nodes.begin(); it2!=node->nodes.end(); it2++) {
		Clear(*it2);
		free((*it2)->name);
		delete *it2;
	}
}

void XML_tree::_open(const char *text)
{
	isOpen=true;
	printf("_open: %s\n",text);
	node_st *node=new node_st;
	node->parent=current;
	node->name=strdup(text);
	current->nodes.push_back(node);
	current=node;
}

void XML_tree::_text(const char *text)
{
	node_st *node=new node_st;
	node->parent=0;
	node->name=strdup(text);
	current->nodes.push_back(node);
	//current=node;
}

void XML_tree::_close(const char * text)
{
	isOpen=false;
	printf("_close: %s\n",text && *text?text:"*");
	if(current==&root) return;
	if(text && *text && strcmp(current->name,text)) return;
	current=current->parent;
}

//! \param name  
//! \param value  
void XML_tree::_attr(const char* text,const char* value)
{
	if(!isOpen) return;
	attr_st *attr=new attr_st;
	attr->name=strdup(text);
	attr->value=strdup(value);
	current->attrs.push_back(attr);
}


void XML_tree::Dump(node_st *node)
{
	for(std::list<attr_st*>::iterator it1=node->attrs.begin(); it1!=node->attrs.end(); it1++) {
		printf(" %s=%s\n",(*it1)->name,(*it1)->value);
	}
	for(std::list<node_st*>::iterator it2=node->nodes.begin(); it2!=node->nodes.end(); it2++) {
		if((*it2)->parent) {
			printf("open: %s\n",(*it2)->name);
			Dump(*it2);
			printf("close: %s\n",(*it2)->name);
		} else {
			printf("text: %s\n",(*it2)->name);
		}
	}
}


void main()
{
/*
	XML_simple xml;
	xml.Parse(
		"</xxx>"
		"<xxx/>"
		"<xxx gh='gg'/>"
		"<ttt rt='jk h\"j' rtt=\"878\" yyy=ngjg >123</ggg>"
		"<!-- ----  > <fff>--!-#--><aaa>"
		);
/*/
	XML_tree xml;
	FILE *file=fopen("calc.xml","rt");
	while(!feof(file)) {
		xml.Parse(fgetc(file));
	}
	fclose(file);
	xml.Dump();
//*/
	getchar();
}



